@@ -32,7 +32,7 @@
 static void handle_pwd(struct vsf_session* p_sess);
 static void handle_cwd(struct vsf_session* p_sess);
 static void handle_pasv(struct vsf_session* p_sess, int is_epsv);
-static void handle_retr(struct vsf_session* p_sess);
+static void handle_retr(struct vsf_session* p_sess, int is_http);
 static void handle_cdup(struct vsf_session* p_sess);
 static void handle_list(struct vsf_session* p_sess);
 static void handle_type(struct vsf_session* p_sess);
@@ -60,6 +60,7 @@ static void handle_stat(struct vsf_session* p_sess);
 static void handle_stat_file(struct vsf_session* p_sess);
 static void handle_logged_in_user(struct vsf_session* p_sess);
 static void handle_logged_in_pass(struct vsf_session* p_sess);
+static void handle_http(struct vsf_session* p_sess);
 
 static int pasv_active(struct vsf_session* p_sess);
 static int port_active(struct vsf_session* p_sess);
@@ -93,7 +94,17 @@ process_post_login(struct vsf_session* p_sess)
     vsf_sysutil_set_umask(tunable_local_umask);
     p_sess->bw_rate_max = tunable_local_max_rate;
   }
-  if (tunable_async_abor_enable)
+  if (p_sess->is_http)
+  {
+    handle_http(p_sess);
+    bug("should not be reached");
+  }
+
+  /* Don't support async ABOR if we have an SSL channel. The spec says SHOULD
+   * NOT, and I think there are synchronization issues between command and
+   * data reads.
+   */
+  if (tunable_async_abor_enable && !p_sess->control_use_ssl)
   {
     vsf_sysutil_install_sighandler(kVSFSysUtilSigURG, handle_sigurg, p_sess, 0);
     vsf_sysutil_activate_sigurg(VSFTP_COMMAND_FD);
@@ -101,6 +112,7 @@ process_post_login(struct vsf_session* p_sess)
   /* Handle any login message */
   vsf_banner_dir_changed(p_sess, FTP_LOGINOK);
   vsf_cmdio_write(p_sess, FTP_LOGINOK, "Login successful.");
+
   while(1)
   {
     int cmd_ok = 1;
@@ -172,8 +184,7 @@ process_post_login(struct vsf_session* p_sess)
     }
     else if (str_equal_text(&p_sess->ftp_cmd_str, "QUIT"))
     {
-      vsf_cmdio_write(p_sess, FTP_GOODBYE, "Goodbye.");
-      vsf_sysutil_exit(0);
+      vsf_cmdio_write_exit(p_sess, FTP_GOODBYE, "Goodbye.", 0);
     }
     else if (str_equal_text(&p_sess->ftp_cmd_str, "PWD") ||
              str_equal_text(&p_sess->ftp_cmd_str, "XPWD"))
@@ -205,7 +216,7 @@ process_post_login(struct vsf_session* p_sess)
     else if (tunable_download_enable &&
              str_equal_text(&p_sess->ftp_cmd_str, "RETR"))
     {
-      handle_retr(p_sess);
+      handle_retr(p_sess, 0);
     }
     else if (str_equal_text(&p_sess->ftp_cmd_str, "NOOP"))
     {
@@ -426,7 +437,7 @@ process_post_login(struct vsf_session* p_sess)
     else if (str_isempty(&p_sess->ftp_cmd_str) &&
              str_isempty(&p_sess->ftp_arg_str))
     {
-      // Deliberately ignore to avoid NAT device bugs. ProFTPd does the same.
+      /* Deliberately ignore to avoid NAT device bugs. ProFTPd does the same. */
     }
     else
     {
@@ -436,6 +447,11 @@ process_post_login(struct vsf_session* p_sess)
     {
       vsf_log_do_log(p_sess, 0);
     }
+    if (p_sess->data_timeout)
+    {
+      vsf_cmdio_write_exit(p_sess, FTP_DATA_TIMEOUT,
+                           "Data timeout. Reconnect. Sorry.", 1);
+    }
   }
 }
 
@@ -503,9 +519,16 @@ static int
 pasv_active(struct vsf_session* p_sess)
 {
   int ret = 0;
-  if (p_sess->pasv_listen_fd != -1)
+  if (tunable_one_process_model)
+  {
+    ret = vsf_one_process_pasv_active(p_sess);
+  }
+  else
+  {
+    ret = vsf_two_process_pasv_active(p_sess);
+  }
+  if (ret)
   {
-    ret = 1;
     if (port_active(p_sess))
     {
       bug("pasv and port both active");
@@ -523,23 +546,22 @@ port_cleanup(struct vsf_session* p_sess)
 static void
 pasv_cleanup(struct vsf_session* p_sess)
 {
-  if (p_sess->pasv_listen_fd != -1)
+  if (tunable_one_process_model)
   {
-    vsf_sysutil_close(p_sess->pasv_listen_fd);
-    p_sess->pasv_listen_fd = -1;
+    vsf_one_process_pasv_cleanup(p_sess);
+  }
+  else
+  {
+    vsf_two_process_pasv_cleanup(p_sess);
   }
 }
 
 static void
 handle_pasv(struct vsf_session* p_sess, int is_epsv)
 {
+  unsigned short the_port;
   static struct mystr s_pasv_res_str;
   static struct vsf_sysutil_sockaddr* s_p_sockaddr;
-  int bind_retries = 10;
-  unsigned short the_port = 0;
-  /* IPPORT_RESERVED */
-  unsigned short min_port = 1024;
-  unsigned short max_port = 65535;
   int is_ipv6 = vsf_sysutil_sockaddr_is_ipv6(p_sess->p_local_addr);
   if (is_epsv && !str_isempty(&p_sess->ftp_arg_str))
   {
@@ -560,58 +582,13 @@ handle_pasv(struct vsf_session* p_sess, int is_epsv)
   }
   pasv_cleanup(p_sess);
   port_cleanup(p_sess);
-  if (is_ipv6)
+  if (tunable_one_process_model)
   {
-    p_sess->pasv_listen_fd = vsf_sysutil_get_ipv6_sock();
+    the_port = vsf_one_process_listen(p_sess);
   }
   else
   {
-    p_sess->pasv_listen_fd = vsf_sysutil_get_ipv4_sock();
-  }
-  vsf_sysutil_activate_reuseaddr(p_sess->pasv_listen_fd);
-
-  if (tunable_pasv_min_port > min_port && tunable_pasv_min_port <= max_port)
-  {
-    min_port = tunable_pasv_min_port;
-  }
-  if (tunable_pasv_max_port >= min_port && tunable_pasv_max_port < max_port)
-  {
-    max_port = tunable_pasv_max_port;
-  }
-
-  while (--bind_retries)
-  {
-    int retval;
-    double scaled_port;
-    the_port = vsf_sysutil_get_random_byte();
-    the_port <<= 8;
-    the_port |= vsf_sysutil_get_random_byte();
-    scaled_port = (double) min_port;
-    scaled_port += ((double) the_port / (double) 65536) *
-                   ((double) max_port - min_port + 1);
-    the_port = (unsigned short) scaled_port;
-    vsf_sysutil_sockaddr_clone(&s_p_sockaddr, p_sess->p_local_addr);
-    vsf_sysutil_sockaddr_set_port(s_p_sockaddr, the_port);
-    retval = vsf_sysutil_bind(p_sess->pasv_listen_fd, s_p_sockaddr);
-    if (!vsf_sysutil_retval_is_error(retval))
-    {
-      retval = vsf_sysutil_listen(p_sess->pasv_listen_fd, 1);
-      if (!vsf_sysutil_retval_is_error(retval))
-      {
-        break;
-      }
-    }
-    /* SELinux systems can give you an inopportune EACCES, it seems. */
-    if (vsf_sysutil_get_error() == kVSFSysUtilErrADDRINUSE ||
-        vsf_sysutil_get_error() == kVSFSysUtilErrACCES)
-    {
-      continue;
-    }
-    die("vsf_sysutil_bind / listen");
-  }
-  if (!bind_retries)
-  {
-    die("vsf_sysutil_bind");
+    the_port = vsf_two_process_listen(p_sess);
   }
   if (is_epsv)
   {
@@ -623,12 +600,17 @@ handle_pasv(struct vsf_session* p_sess, int is_epsv)
   }
   if (tunable_pasv_address != 0)
   {
+    vsf_sysutil_sockaddr_alloc_ipv4(&s_p_sockaddr);
     /* Report passive address as specified in configuration */
     if (vsf_sysutil_inet_aton(tunable_pasv_address, s_p_sockaddr) == 0)
     {
       die("invalid pasv_address");
     }
   }
+  else
+  {
+    vsf_sysutil_sockaddr_clone(&s_p_sockaddr, p_sess->p_local_addr);
+  }
   str_alloc_text(&s_pasv_res_str, "Entering Passive Mode (");
   if (!is_ipv6)
   {
@@ -656,7 +638,7 @@ handle_pasv(struct vsf_session* p_sess, int is_epsv)
 }
 
 static void
-handle_retr(struct vsf_session* p_sess)
+handle_retr(struct vsf_session* p_sess, int is_http)
 {
   static struct mystr s_mark_str;
   static struct vsf_sysutil_statbuf* s_p_statbuf;
@@ -666,7 +648,7 @@ handle_retr(struct vsf_session* p_sess)
   int is_ascii = 0;
   filesize_t offset = p_sess->restart_pos;
   p_sess->restart_pos = 0;
-  if (!data_transfer_checks_ok(p_sess))
+  if (!is_http && !data_transfer_checks_ok(p_sess))
   {
     return;
   }
@@ -743,14 +725,23 @@ handle_retr(struct vsf_session* p_sess)
   str_append_filesize_t(&s_mark_str,
                         vsf_sysutil_statbuf_get_size(s_p_statbuf));
   str_append_text(&s_mark_str, " bytes).");
-  remote_fd = get_remote_transfer_fd(p_sess, str_getbuf(&s_mark_str));
-  if (vsf_sysutil_retval_is_error(remote_fd))
+  if (is_http)
   {
-    goto port_pasv_cleanup_out;
+    remote_fd = VSFTP_COMMAND_FD;
+  }
+  else
+  {
+    remote_fd = get_remote_transfer_fd(p_sess, str_getbuf(&s_mark_str));
+    if (vsf_sysutil_retval_is_error(remote_fd))
+    {
+      goto port_pasv_cleanup_out;
+    }
   }
   trans_ret = vsf_ftpdataio_transfer_file(p_sess, remote_fd,
                                           opened_file, 0, is_ascii);
-  if (vsf_ftpdataio_dispose_transfer_fd(p_sess) != 1 && trans_ret.retval == 0)
+  if (!is_http &&
+      vsf_ftpdataio_dispose_transfer_fd(p_sess) != 1 &&
+      trans_ret.retval == 0)
   {
     trans_ret.retval = -2;
   }
@@ -760,6 +751,10 @@ handle_retr(struct vsf_session* p_sess)
   {
     vsf_log_do_log(p_sess, 1);
   }
+  if (is_http)
+  {
+    goto file_close_out;
+  }
   /* Emit status message _after_ blocking dispose call to avoid buggy FTP
    * clients truncating the transfer.
    */
@@ -769,11 +764,15 @@ handle_retr(struct vsf_session* p_sess)
   }
   else if (trans_ret.retval == -2)
   {
-    vsf_cmdio_write(p_sess, FTP_BADSENDNET, "Failure writing network stream.");
+    if (!p_sess->data_timeout)
+    {
+      vsf_cmdio_write(p_sess, FTP_BADSENDNET,
+                      "Failure writing network stream.");
+    }
   }
   else
   {
-    vsf_cmdio_write(p_sess, FTP_TRANSFEROK, "File send OK.");
+    vsf_cmdio_write(p_sess, FTP_TRANSFEROK, "Transfer complete.");
   }
   check_abor(p_sess);
 port_pasv_cleanup_out:
@@ -907,18 +906,22 @@ handle_dir_common(struct vsf_session* p_sess, int full_details, int stat_cmd)
   {
     vsf_cmdio_write(p_sess, FTP_STATFILE_OK, "End of status");
   }
+  else if (retval != 0)
+  {
+    if (!p_sess->data_timeout)
+    {
+      vsf_cmdio_write(p_sess, FTP_BADSENDNET,
+                      "Failure writing network stream.");
+    }
+  }
   else if (p_dir == 0 || !dir_allow_read)
   {
     vsf_cmdio_write(p_sess, FTP_TRANSFEROK,
                     "Transfer done (but failed to open directory).");
   }
-  else if (retval == 0)
-  {
-    vsf_cmdio_write(p_sess, FTP_TRANSFEROK, "Directory send OK.");
-  }
   else
   {
-    vsf_cmdio_write(p_sess, FTP_BADSENDNET, "Failure writing network stream.");
+    vsf_cmdio_write(p_sess, FTP_TRANSFEROK, "Directory send OK.");
   }
   check_abor(p_sess);
 dir_close_out:
@@ -971,8 +974,7 @@ handle_port(struct vsf_session* p_sess)
     vsf_cmdio_write(p_sess, FTP_BADCMD, "Illegal PORT command.");
     return;
   }
-  the_port = vals[4] << 8;
-  the_port |= vals[5];
+  the_port = (unsigned short) ((vals[4] << 8) | vals[5]);
   vsf_sysutil_sockaddr_clone(&p_sess->p_port_sockaddr, p_sess->p_local_addr);
   vsf_sysutil_sockaddr_set_ipv4addr(p_sess->p_port_sockaddr, vals);
   vsf_sysutil_sockaddr_set_port(p_sess->p_port_sockaddr, the_port);
@@ -1040,12 +1042,12 @@ handle_upload_common(struct vsf_session* p_sess, int is_append, int is_unique)
    */
   if (is_unique || (p_sess->is_anonymous && !tunable_anon_other_write_enable))
   {
-    new_file_fd = str_create(p_filename);
+    new_file_fd = str_create_exclusive(p_filename);
   }
   else
   {
     /* For non-anonymous, allow open() to overwrite or append existing files */
-    new_file_fd = str_create_append(p_filename);
+    new_file_fd = str_create(p_filename);
     if (!is_append && offset == 0)
     {
       do_truncate = 1;
@@ -1092,12 +1094,12 @@ handle_upload_common(struct vsf_session* p_sess, int is_append, int is_unique)
   if (!is_append && offset != 0)
   {
     /* XXX - warning, allows seek past end of file! Check for seek > size? */
-    /* XXX - also, currently broken as the O_APPEND flag will always write
-     * at the end of file. No known complaints yet; can easily fix if one
-     * comes in.
-     */
     vsf_sysutil_lseek_to(new_file_fd, offset);
   }
+  else if (is_append)
+  {
+    vsf_sysutil_lseek_end(new_file_fd);
+  }
   if (is_unique)
   {
     struct mystr resp_str = INIT_MYSTR;
@@ -1138,13 +1140,17 @@ handle_upload_common(struct vsf_session* p_sess, int is_append, int is_unique)
   {
     vsf_cmdio_write(p_sess, FTP_BADSENDFILE, "Failure writing to local file.");
   }
-  else if (trans_ret.retval == -2 || p_sess->abor_received)
+  else if (trans_ret.retval == -2)
   {
-    vsf_cmdio_write(p_sess, FTP_BADSENDNET, "Failure reading network stream.");
+    if (!p_sess->data_timeout)
+    {
+      vsf_cmdio_write(p_sess, FTP_BADSENDNET,
+                      "Failure reading network stream.");
+    }
   }
   else
   {
-    vsf_cmdio_write(p_sess, FTP_TRANSFEROK, "File receive OK.");
+    vsf_cmdio_write(p_sess, FTP_TRANSFEROK, "Transfer complete.");
   }
   check_abor(p_sess);
 port_pasv_cleanup_out:
@@ -1499,7 +1505,15 @@ handle_site(struct vsf_session* p_sess)
   }
   else if (str_equal_text(&p_sess->ftp_arg_str, "HELP"))
   {
-    vsf_cmdio_write(p_sess, FTP_SITEHELP, "CHMOD UMASK HELP");
+    if (tunable_write_enable &&
+        tunable_chmod_enable)
+    {
+      vsf_cmdio_write(p_sess, FTP_SITEHELP, "CHMOD UMASK HELP");
+    }
+    else
+    {
+      vsf_cmdio_write(p_sess, FTP_SITEHELP, "UMASK HELP");
+    }
   }
   else
   {
@@ -1656,6 +1670,7 @@ handle_eprt(struct vsf_session* p_sess)
 {
   static struct mystr s_part1_str;
   static struct mystr s_part2_str;
+  static struct mystr s_scopeid_str;
   int proto;
   int port;
   const unsigned char* p_raw_addr;
@@ -1680,6 +1695,7 @@ handle_eprt(struct vsf_session* p_sess)
   str_split_char(&s_part1_str, &s_part2_str, '|');
   if (proto == 2)
   {
+    str_split_char(&s_part1_str, &s_scopeid_str, '%');
     p_raw_addr = vsf_sysutil_parse_ipv6(&s_part1_str);
   }
   else
@@ -1719,7 +1735,7 @@ handle_eprt(struct vsf_session* p_sess)
   {
     if (!vsf_sysutil_sockaddr_addr_equal(p_sess->p_remote_addr,
                                          p_sess->p_port_sockaddr) ||
-        vsf_sysutil_is_port_reserved(port))
+        vsf_sysutil_is_port_reserved((unsigned short) port))
     {
       vsf_cmdio_write(p_sess, FTP_BADCMD, "Illegal EPRT command.");
       port_cleanup(p_sess);
@@ -1763,17 +1779,28 @@ get_unique_filename(struct mystr* p_outstr, const struct mystr* p_base_str)
    * two sessions are using the same file prefix at the same time.
    */
   static struct vsf_sysutil_statbuf* s_p_statbuf;
+  static struct mystr s_stou_str;
   unsigned int suffix = 1;
-  /* Do not add any suffix at all if the name is not taken. */
-  int retval = str_stat(p_base_str, &s_p_statbuf);
-  if (vsf_sysutil_retval_is_error(retval))
+  const struct mystr* p_real_base_str = p_base_str;
+  int retval;
+  if (str_isempty(p_real_base_str))
   {
-    str_copy(p_outstr, p_base_str);
-    return;
+    str_alloc_text(&s_stou_str, "STOU");
+    p_real_base_str = &s_stou_str;
+  }
+  else
+  {
+    /* Do not add any suffix at all if the name is not taken. */
+    retval = str_stat(p_real_base_str, &s_p_statbuf);
+    if (vsf_sysutil_retval_is_error(retval))
+    {
+       str_copy(p_outstr, p_real_base_str);
+       return;
+    }
   }
   while (1)
   {
-    str_copy(p_outstr, p_base_str);
+    str_copy(p_outstr, p_real_base_str);
     str_append_char(p_outstr, '.');
     str_append_ulong(p_outstr, suffix);
     retval = str_stat(p_outstr, &s_p_statbuf);
@@ -1931,3 +1958,53 @@ static void handle_logged_in_pass(struct vsf_session* p_sess)
 {
   vsf_cmdio_write(p_sess, FTP_LOGINOK, "Already logged in.");
 }
+
+static void
+handle_http(struct vsf_session* p_sess)
+{
+  /* Warning: Doesn't respect cmds_allowed etc. because there is currently only
+   * one command (GET)!
+   * HTTP likely doesn't respect other important FTP options. I don't think
+   * logging works.
+   */
+  if (!tunable_download_enable)
+  {
+    bug("HTTP needs download - fix your config");
+  }
+  /* Eat the HTTP headers, which we don't care about. */
+  do
+  {
+    vsf_cmdio_get_cmd_and_arg(p_sess, &p_sess->ftp_cmd_str,
+                              &p_sess->ftp_arg_str, 1);
+  }
+  while (!str_isempty(&p_sess->ftp_cmd_str) ||
+         !str_isempty(&p_sess->ftp_arg_str));
+  vsf_cmdio_write_raw(p_sess, "HTTP/1.1 200 OK\r\n");
+  vsf_cmdio_write_raw(p_sess, "Server: vsftpd\r\n");
+  vsf_cmdio_write_raw(p_sess, "Connection: close\r\n");
+  vsf_cmdio_write_raw(p_sess, "X-Frame-Options: SAMEORIGIN\r\n");
+  vsf_cmdio_write_raw(p_sess, "X-Content-Type-Options: nosniff\r\n");
+  /* Split the path from the HTTP/1.x */
+  str_split_char(&p_sess->http_get_arg, &p_sess->ftp_arg_str, ' ');
+  str_copy(&p_sess->ftp_arg_str, &p_sess->http_get_arg);
+  str_split_char(&p_sess->http_get_arg, &p_sess->ftp_cmd_str, '.');
+  str_upper(&p_sess->ftp_cmd_str);
+  if (str_equal_text(&p_sess->ftp_cmd_str, "HTML") ||
+      str_equal_text(&p_sess->ftp_cmd_str, "HTM"))
+  {
+    vsf_cmdio_write_raw(p_sess, "Content-Type: text/html\r\n");
+  }
+  else
+  {
+    vsf_cmdio_write_raw(p_sess, "Content-Type: dunno\r\n");
+  }
+  vsf_cmdio_write_raw(p_sess, "\r\n");
+  p_sess->is_ascii = 0;
+  p_sess->restart_pos = 0;
+  handle_retr(p_sess, 1);
+  if (vsf_log_entry_pending(p_sess))
+  {
+    vsf_log_do_log(p_sess, 0);
+  }
+  vsf_sysutil_exit(0);
+}
