@@ -73,7 +73,7 @@ static struct vsf_sysutil_sig_details
 {
   vsf_sighandle_t sync_sig_handler;
   void* p_private;
-  int pending;
+  volatile sig_atomic_t pending;
   int running;
   int use_alarm;
 } s_sig_details[NSIG];
@@ -117,9 +117,8 @@ vsf_sysutil_common_sighandler(int signum)
 {
   if (signum < 0 || signum >= NSIG)
   {
-    // bug() is not async safe but this check really is a "cannot happen"
-    // debug aid.
-    bug("signal out of range in vsf_sysutil_common_sighandler");
+    /* "cannot happen" */
+    return;
   }
   if (s_sig_details[signum].sync_sig_handler)
   {
@@ -448,7 +447,7 @@ vsf_sysutil_write_loop(const int fd, const void* p_buf, unsigned int size)
     }
     if ((unsigned int) retval > size)
     {
-      die("retval too big in vsf_sysutil_read_loop");
+      die("retval too big in vsf_sysutil_write_loop");
     }
     num_written += retval;
     size -= (unsigned int) retval;
@@ -486,6 +485,17 @@ vsf_sysutil_lseek_to(const int fd, filesize_t seek_pos)
   }
 }
 
+void
+vsf_sysutil_lseek_end(const int fd)
+{
+  filesize_t retval;
+  retval = lseek(fd, 0, SEEK_END);
+  if (retval < 0)
+  {
+    die("lseek");
+  }
+}
+
 void*
 vsf_sysutil_malloc(unsigned int size)
 {
@@ -542,15 +552,7 @@ vsf_sysutil_getpid(void)
 int
 vsf_sysutil_fork(void)
 {
-  /* Child does NOT inherit exit function */
-  exitfunc_t curr_func = s_exit_func;
-  int retval;
-  s_exit_func = 0;
-  retval = vsf_sysutil_fork_failok();
-  if (retval != 0)
-  {
-    s_exit_func = curr_func;
-  }
+  int retval = vsf_sysutil_fork_failok();
   if (retval < 0)
   {
     die("fork");
@@ -561,10 +563,11 @@ vsf_sysutil_fork(void)
 int
 vsf_sysutil_fork_failok(void)
 {
-  int retval = fork();
+  int retval;
+  retval = fork();
   if (retval == 0)
   {
-    vsf_sysutil_clear_pid_cache();
+    vsf_sysutil_post_fork();
   }
   return retval;
 }
@@ -722,7 +725,7 @@ vsf_sysutil_activate_linger(int fd)
   struct linger the_linger;
   vsf_sysutil_memclr(&the_linger, sizeof(the_linger));
   the_linger.l_onoff = 1;
-  the_linger.l_linger = 32767;
+  the_linger.l_linger = 60 * 10;
   retval = setsockopt(fd, SOL_SOCKET, SO_LINGER, &the_linger,
                       sizeof(the_linger));
   if (retval != 0)
@@ -906,13 +909,13 @@ vsf_sysutil_octal_to_uint(const char* p_str)
 int
 vsf_sysutil_toupper(int the_char)
 {
-  return toupper(the_char);
+  return toupper((unsigned char) the_char);
 }
 
 int
 vsf_sysutil_isspace(int the_char)
 {
-  return isspace(the_char);
+  return isspace((unsigned char) the_char);
 }
 
 int
@@ -940,13 +943,13 @@ vsf_sysutil_isprint(int the_char)
 int
 vsf_sysutil_isalnum(int the_char)
 {
-  return isalnum(the_char);
+  return isalnum((unsigned char) the_char);
 }
 
 int
 vsf_sysutil_isdigit(int the_char)
 {
-  return isdigit(the_char);
+  return isdigit((unsigned char) the_char);
 }
 
 char*
@@ -1017,13 +1020,13 @@ vsf_sysutil_next_dirent(struct vsf_sysutil_dir* p_dir)
 unsigned int
 vsf_sysutil_strlen(const char* p_text)
 {
-  unsigned int ret = strlen(p_text);
+  size_t ret = strlen(p_text);
   /* A defense in depth measure. */
   if (ret > INT_MAX / 8)
   {
     die("string suspiciously long");
   }
-  return ret;
+  return (unsigned int) ret;
 }
 
 char*
@@ -1173,7 +1176,7 @@ vsf_sysutil_open_file(const char* p_filename,
 }
 
 int
-vsf_sysutil_create_file(const char* p_filename)
+vsf_sysutil_create_file_exclusive(const char* p_filename)
 {
   /* umask() also contributes to end mode */
   return open(p_filename, O_CREAT | O_EXCL | O_WRONLY | O_APPEND,
@@ -1181,17 +1184,16 @@ vsf_sysutil_create_file(const char* p_filename)
 }
 
 int
-vsf_sysutil_create_overwrite_file(const char* p_filename)
+vsf_sysutil_create_or_open_file(const char* p_filename, unsigned int mode)
 {
-  return open(p_filename, O_CREAT | O_TRUNC | O_WRONLY |
-                          O_APPEND | O_NONBLOCK,
-              tunable_file_open_mode);
+  return open(p_filename, O_CREAT | O_WRONLY | O_NONBLOCK, mode);
 }
 
 int
-vsf_sysutil_create_or_open_file(const char* p_filename, unsigned int mode)
+vsf_sysutil_create_or_open_file_append(const char* p_filename,
+                                       unsigned int mode)
 {
-  return open(p_filename, O_CREAT | O_WRONLY | O_APPEND | O_NONBLOCK, mode);
+  return open(p_filename, O_CREAT | O_WRONLY | O_NONBLOCK | O_APPEND, mode);
 }
 
 void
@@ -1338,6 +1340,7 @@ vsf_sysutil_statbuf_get_perms(const struct vsf_sysutil_statbuf* p_statbuf)
     case S_IFSOCK: perms[0] = 's'; break;
     case S_IFCHR: perms[0] = 'c'; break;
     case S_IFBLK: perms[0] = 'b'; break;
+    default: break;
   }
   if (p_stat->st_mode & S_IRUSR) perms[1] = 'r';
   if (p_stat->st_mode & S_IWUSR) perms[2] = 'w';
@@ -1357,13 +1360,12 @@ vsf_sysutil_statbuf_get_perms(const struct vsf_sysutil_statbuf* p_statbuf)
 
 const char*
 vsf_sysutil_statbuf_get_date(const struct vsf_sysutil_statbuf* p_statbuf,
-                             int use_localtime)
+                             int use_localtime, long curr_time)
 {
   static char datebuf[64];
   int retval;
   struct tm* p_tm;
   const struct stat* p_stat = (const struct stat*) p_statbuf;
-  long local_time = vsf_sysutil_get_cached_time_sec();
   const char* p_date_format = "%b %d %H:%M";
   if (!use_localtime)
   {
@@ -1374,8 +1376,8 @@ vsf_sysutil_statbuf_get_date(const struct vsf_sysutil_statbuf* p_statbuf,
     p_tm = localtime(&p_stat->st_mtime);
   }
   /* Is this a future or 6 months old date? If so, we drop to year format */
-  if (p_stat->st_mtime > local_time ||
-      (local_time - p_stat->st_mtime) > 60*60*24*182)
+  if (p_stat->st_mtime > curr_time ||
+      (curr_time - p_stat->st_mtime) > 60*60*24*182)
   {
     p_date_format = "%b %d  %Y";
   }
@@ -1597,7 +1599,12 @@ vsf_sysutil_get_error(void)
       retval = kVSFSysUtilErrOPNOTSUPP;
       break;
     case EACCES:
-      retval = kVSFSysUtilErrOPNOTSUPP;
+      retval = kVSFSysUtilErrACCES;
+      break;
+    case ENOENT:
+      retval = kVSFSysUtilErrNOENT;
+      break;
+    default:
       break;
   }
   return retval;
@@ -1684,7 +1691,7 @@ vsf_sysutil_accept_timeout(int fd, struct vsf_sysutil_sockaddr* p_sockaddr,
   int saved_errno;
   fd_set accept_fdset;
   struct timeval timeout;
-  unsigned int socklen = sizeof(remote_addr);
+  socklen_t socklen = sizeof(remote_addr);
   if (p_sockaddr)
   {
     vsf_sysutil_memclr(p_sockaddr, sizeof(*p_sockaddr));
@@ -1700,10 +1707,14 @@ vsf_sysutil_accept_timeout(int fd, struct vsf_sysutil_sockaddr* p_sockaddr,
       retval = select(fd + 1, &accept_fdset, NULL, NULL, &timeout);
       saved_errno = errno;
       vsf_sysutil_check_pending_actions(kVSFSysUtilUnknown, 0, 0);
-    } while (retval < 0 && saved_errno == EINTR);
-    if (retval == 0)
+    }
+    while (retval < 0 && saved_errno == EINTR);
+    if (retval <= 0)
     {
-      errno = EAGAIN;
+      if (retval == 0)
+      {
+        errno = EAGAIN;
+      }
       return -1;
     }
   }
@@ -1781,10 +1792,13 @@ vsf_sysutil_connect_timeout(int fd, const struct vsf_sysutil_sockaddr* p_addr,
       vsf_sysutil_check_pending_actions(kVSFSysUtilUnknown, 0, 0);
     }
     while (retval < 0 && saved_errno == EINTR);
-    if (retval == 0)
+    if (retval <= 0)
     {
+      if (retval == 0)
+      {
+        errno = EAGAIN;
+      }
       retval = -1;
-      errno = EAGAIN;
     }
     else
     {
@@ -1794,6 +1808,11 @@ vsf_sysutil_connect_timeout(int fd, const struct vsf_sysutil_sockaddr* p_addr,
       {
         die("getsockopt");
       }
+      if (retval != 0)
+      {
+        errno = retval;
+        retval = -1;
+      }
     }
   }
   if (wait_seconds > 0)
@@ -1808,7 +1827,7 @@ vsf_sysutil_getsockname(int fd, struct vsf_sysutil_sockaddr** p_sockptr)
 {
   struct vsf_sysutil_sockaddr the_addr;
   int retval;
-  unsigned int socklen = sizeof(the_addr);
+  socklen_t socklen = sizeof(the_addr);
   vsf_sysutil_sockaddr_clear(p_sockptr);
   retval = getsockname(fd, &the_addr.u.u_sockaddr, &socklen);
   if (retval != 0)
@@ -1833,7 +1852,7 @@ vsf_sysutil_getpeername(int fd, struct vsf_sysutil_sockaddr** p_sockptr)
 {
   struct vsf_sysutil_sockaddr the_addr;
   int retval;
-  unsigned int socklen = sizeof(the_addr);
+  socklen_t socklen = sizeof(the_addr);
   vsf_sysutil_sockaddr_clear(p_sockptr);
   retval = getpeername(fd, &the_addr.u.u_sockaddr, &socklen);
   if (retval != 0)
@@ -1925,6 +1944,8 @@ vsf_sysutil_sockaddr_clone(struct vsf_sysutil_sockaddr** p_sockptr,
     vsf_sysutil_memcpy(&p_sockaddr->u.u_sockaddr_in6.sin6_addr,
                        &p_src->u.u_sockaddr_in6.sin6_addr,
                        sizeof(p_sockaddr->u.u_sockaddr_in6.sin6_addr));
+    p_sockaddr->u.u_sockaddr_in6.sin6_scope_id =
+        p_src->u.u_sockaddr_in6.sin6_scope_id;
   }
   else
   {
@@ -2007,7 +2028,7 @@ vsf_sysutil_sockaddr_set_ipv4addr(struct vsf_sysutil_sockaddr* p_sockptr,
     static struct vsf_sysutil_sockaddr* s_p_sockaddr;
     vsf_sysutil_sockaddr_alloc_ipv4(&s_p_sockaddr);
     vsf_sysutil_memcpy(&s_p_sockaddr->u.u_sockaddr_in.sin_addr, p_raw,
-                       sizeof(&s_p_sockaddr->u.u_sockaddr_in.sin_addr));
+                       sizeof(s_p_sockaddr->u.u_sockaddr_in.sin_addr));
     vsf_sysutil_memcpy(&p_sockptr->u.u_sockaddr_in6.sin6_addr,
                        vsf_sysutil_sockaddr_ipv4_v6(s_p_sockaddr),
                        sizeof(p_sockptr->u.u_sockaddr_in6.sin6_addr));
@@ -2036,7 +2057,8 @@ vsf_sysutil_sockaddr_set_ipv6addr(struct vsf_sysutil_sockaddr* p_sockptr,
 const void*
 vsf_sysutil_sockaddr_ipv6_v4(const struct vsf_sysutil_sockaddr* p_addr)
 {
-  static char pattern[12] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0xFF, 0xFF };
+  static unsigned char pattern[12] =
+      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0xFF, 0xFF };
   const unsigned char* p_addr_start;
   if (p_addr->u.u_sockaddr.sa_family != AF_INET6)
   {
@@ -2053,7 +2075,7 @@ vsf_sysutil_sockaddr_ipv6_v4(const struct vsf_sysutil_sockaddr* p_addr)
 const void*
 vsf_sysutil_sockaddr_ipv4_v6(const struct vsf_sysutil_sockaddr* p_addr)
 {
-  static char ret[16] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0xFF, 0xFF };
+  static unsigned char ret[16] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0xFF, 0xFF };
   if (p_addr->u.u_sockaddr.sa_family != AF_INET)
   {
     return 0;
@@ -2130,6 +2152,25 @@ vsf_sysutil_sockaddr_set_any(struct vsf_sysutil_sockaddr* p_sockaddr)
   }
 }
 
+unsigned short
+vsf_sysutil_sockaddr_get_port(const struct vsf_sysutil_sockaddr* p_sockptr)
+{
+  if (p_sockptr->u.u_sockaddr.sa_family == AF_INET)
+  {
+    return ntohs(p_sockptr->u.u_sockaddr_in.sin_port);
+  }
+  else if (p_sockptr->u.u_sockaddr.sa_family == AF_INET6)
+  {
+    return ntohs(p_sockptr->u.u_sockaddr_in6.sin6_port);
+  }
+  else
+  {
+    bug("bad family");
+  }
+  /* NOTREACHED */
+  return 0;
+}
+
 void
 vsf_sysutil_sockaddr_set_port(struct vsf_sysutil_sockaddr* p_sockptr,
                               unsigned short the_port)
@@ -2559,6 +2600,19 @@ vsf_sysutil_tzset(void)
       s_timezone *= -1;
     }
   }
+  /* Call in to the time subsystem again now that TZ is set, trying to force
+   * caching of the actual zoneinfo for the timezone.
+   */
+  p_tm = localtime(&the_time);
+  if (p_tm == NULL)
+  {
+    die("localtime #2");
+  }
+  p_tm = gmtime(&the_time);
+  if (p_tm == NULL)
+  {
+    die("gmtime");
+  }
 }
 
 const char*
@@ -2568,8 +2622,7 @@ vsf_sysutil_get_current_date(void)
   time_t curr_time;
   const struct tm* p_tm;
   int i = 0;
-  vsf_sysutil_update_cached_time();
-  curr_time = vsf_sysutil_get_cached_time_sec();
+  curr_time = vsf_sysutil_get_time_sec();
   p_tm = localtime(&curr_time);
   if (strftime(datebuf, sizeof(datebuf), "%a %b!%d %H:%M:%S %Y", p_tm) == 0)
   {
@@ -2592,23 +2645,18 @@ vsf_sysutil_get_current_date(void)
   return datebuf;
 }
 
-void
-vsf_sysutil_update_cached_time(void)
+long
+vsf_sysutil_get_time_sec(void)
 {
   if (gettimeofday(&s_current_time, NULL) != 0)
   {
     die("gettimeofday");
   }
-}
-
-long
-vsf_sysutil_get_cached_time_sec(void)
-{
   return s_current_time.tv_sec;
 }
 
 long
-vsf_sysutil_get_cached_time_usec(void)
+vsf_sysutil_get_time_usec(void)
 {
   return s_current_time.tv_usec;
 }
@@ -2742,7 +2790,7 @@ vsf_sysutil_getuid(void)
 }
 
 void
-vsf_sysutil_set_address_space_limit(long bytes)
+vsf_sysutil_set_address_space_limit(unsigned long bytes)
 {
   /* Unfortunately, OpenBSD is missing RLIMIT_AS. */
 #ifdef RLIMIT_AS
@@ -2779,6 +2827,7 @@ vsf_sysutil_set_no_fds()
 void
 vsf_sysutil_set_no_procs()
 {
+#ifdef RLIMIT_NPROC
   int ret;
   struct rlimit rlim;
   rlim.rlim_cur = 0;
@@ -2788,10 +2837,25 @@ vsf_sysutil_set_no_procs()
   {
     die("setrlimit NPROC");
   }
+#endif
 }
 
 void
-vsf_sysutil_clear_pid_cache()
+vsf_sysutil_post_fork()
 {
+  int i;
+  /* Don't inherit any exit function. */
+  s_exit_func = NULL;
+  /* Uncached the current PID. */ 
   s_current_pid = -1;
+  /* Don't inherit anything relating to the synchronous signal system */
+  s_io_handler = NULL;
+  for (i=0; i < NSIG; ++i)
+  {
+    s_sig_details[i].sync_sig_handler = NULL;
+  }
+  for (i=0; i < NSIG; ++i)
+  {
+    s_sig_details[i].pending = 0;
+  }
 }
